"""auxjad
======

``auxjad`` is a library of auxiliary functions and classes for Abjad 3.1. All
classes and functions have a __doc__ attribute with usage instructions.

Documentation is available at https://gilbertohasnofb.github.io/auxjad-docs/. A
pdf version of the documentation is also available in the ``/docs`` directory
of the repository.

Bugs can be reported through the project's Issue Tracker.

This library is published under the MIT License.
"""

from .core import *
from .entities import *
from .utilities import *


__author__ = "Gilberto Agostinho <gilbertohasnofb@gmail.com>"
__version__ = "0.7.0"
